import jDisco.*;
import jDisco.Process;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/*
 * This program animates a number of balls moving inside a rectangular box.
 * 
 * A ball reverses direction if it touches any of the four sides of the box. 
 * If two balls collide they bounce off each other.
 *
 * A Ball-object represents a ball and its associated continuous motion. Each ball 
 * has a position (x, y), velocities in the x and y directions (vx, vy), a radius 
 * and a color. The continuous motion of a ball is specified in its derivatives method.
 *
 * The classes BounceHandler and CollisionHandler handle the bouncing events.
 *
 * For each ball there is a BounceHandler-object. Each time a ball touches any of 
 * the sides of the box its direction is reversed.
 *
 * An object of class CollisionHandler handles all collisions between balls. 
 * Anytime two balls collide their velocities are exchanged.
 */

public class BouncingBalls extends Process {
    static Vector allBalls = new Vector();

    class Ball extends Continuous {
        Variable x, y;
        double radius, vx, vy;
        Color color;

        Ball(double x0, double y0, double radius,
             double vx, double vy, Color color) {
            x = new Variable(x0).start();
            y = new Variable(y0).start();
            this.radius = radius; this.color = color;
            this.vx = vx; this.vy = vy;
            allBalls.addElement(this);
            start();
        }

        public void derivatives() {
            x.rate = vx;
            y.rate = vy;
        }

        public double distance(Ball b) {
            double dx = x.state - b.x.state;
            double dy = y.state - b.y.state;
            return Math.sqrt(dx * dx + dy * dy);
        }

        public boolean collidesWith(Ball b) {
            return b != this && distance(b) < radius + b.radius;
        }

        public void draw(Graphics g) {
            Color oldColor = g.getColor();
            g.setColor(color);
            g.fillOval((int) (x.state - radius), 
                       (int) (y.state - radius), 
                       (int) (radius * 2),
                       (int) (radius * 2));
            g.setColor(oldColor);
        }
    }
    
    class BounceHandler extends Process {
        BounceHandler(Ball ball, Dimension d) {
            this.ball = ball;
            this.d = d;
        }

        boolean horizontalBounce() {
            return ball.x.state < ball.radius ||
                   ball.x.state > d.width - ball.radius;
        }

        boolean verticalBounce() {
            return ball.y.state < ball.radius ||
                   ball.y.state > d.height - ball.radius;
        }

        boolean bounce() {
            return horizontalBounce() || verticalBounce();
        }

        public void actions() {
            while (true) {
                waitUntil(new Condition() {
                    public boolean test() {
                        return bounce();
                    }
                });
                if (horizontalBounce()) {
                    ball.vx = -ball.vx;
                    waitUntil(new Condition() {
                        public boolean test() {
                            return !horizontalBounce();
                        }
                    });
                } else if (verticalBounce()) {
                    ball.vy = -ball.vy;
                    waitUntil(new Condition() {
                        public boolean test() {
                            return !verticalBounce();
                        }
                    });
                }
            }
       }
  
       Ball ball;
       Dimension d;
   }
     
   class CollisionHandler extends Process {
        boolean collision() {
            hit = null;
            for (int i = 0; i < allBalls.size(); i++) {
                ball = (Ball) allBalls.elementAt(i);
                for (int j = i + 1; j < allBalls.size();
                     j++) {
                  hit = (Ball) allBalls.elementAt(j);
                  if (ball.collidesWith(hit))
                      return true;
                }
            }
            ball = hit = null;
            return false;
        }

        public void actions() {
            while (true) {
                waitUntil(new Condition() {
                    public boolean test() {
                        return collision();
                    }
                });
                double t;
                t = ball.vx; ball.vx = hit.vx; hit.vx = t;
                t = ball.vy; ball.vy = hit.vy; hit.vy = t;
                waitUntil(new Condition() {
                    public boolean test() {
                        return !collision();
                    }
                });
            }
        }

        Ball ball, hit;
    }

    class Canvas extends JComponent {
        public void paintComponent(Graphics g) {
            g.setColor(Color.white);
            Dimension d = getSize();
            g.fillRect(0, 0, d.width, d.height);
            for (int i = 0; i < allBalls.size(); i++) {
                Ball b = (Ball) allBalls.elementAt(i);
                b.draw(g);
            }
        }
    }

    public void actions() {
        dtMin = 1.0e-5; dtMax = 1; maxRelError = 1.0e-5;
        final JFrame frame = new JFrame();
        frame.getContentPane().add(new Canvas());
        frame.setSize(250, 150);
        frame.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        }); 
        Dimension d = frame.getSize();
        new Ball(d.width * 2 / 3, d.height - 20, 20, -2, -4, 
                 Color.green);
        new Ball(    d.width / 4, d.height - 20, 20,  2, 10, 
                 Color.red);
        new Ball(   d.width / 10, d.height - 40, 20, -5,  3,
                 Color.blue);
        for (int i = 0; i < allBalls.size(); i++)
            activate(new BounceHandler(
                         (Ball) allBalls.elementAt(i), d));
        activate(new CollisionHandler());
        frame.setVisible(true);
        activate(new Process() {
            public void actions() {
                while (true) {
                    frame.repaint();
                    hold(0.1);
                }
            }
        });
        hold(10000);
    }

    public static void main(String[] args) {
        activate(new BouncingBalls());
    }
}
