import jDisco.*;
import jDisco.Process;

/*
 * This example was suggested by F. E. Cellier as a be benchmark problem 
 * that can be used to test the capability of a variable-structure simulation, 
 * that is, a simulation in which the number of differential equations varies 
 * with time.
 *
 * Fifty-five identical stones of the Domino game are placed vertically upright 
 * in a sequence with the same distance between any two stones. If the first 
 * stone is pushed, a chain reaction is started and all stones fall flat.
 *
 * The aim of the simulation is to determine the distance between successive stones 
 * that maximizes the velocity of the chain.
 *
 * A precise description of the model and the experiment is given in the program. 
 * The program illustrates how the specification of the model (DominoGame) can be 
 * separated from the experiment (main).
 *
 * A Stone-object represents a falling stone and the associated discrete events: 
 * the pushing of the next stone and the termination of the fall.
 *
 * Newton's law governs the fall. For each stone we have the equation
 *
 *     theta * phi'' =  m * g * r * sin(phi)
 *
 * where theta is the moment of inertia, hpi is the inclination, m the mass, 
 * g the acceleration, and r is half the diagonal . 
 *
 * In class StoneFall this second-order differential equation is formulated by means 
 * of two first-order differential equations.
 *
 * Each falling stone is associated with one StoneFall-object. A stone that is 
 * not moving - either because it has not yet been pushed, or because it has 
 * already fallen and lies still - is not associated with such an object. 
 * Thus, the number of differential equations will vary with time.     
 *
 * The method maximum determines a maximum of the function y = f(x) on the 
 * interval [a, b]. The maximum is found within the specified tolerance, tol, 
 * by using the golden section search method.
 *
 * Reference:
 * 
 *    Cellier, C. F.
 *    Combined Continuous Discrete Simulation by Use of Digital Computers: 
 *    Techniques and Tools
 *    PhD thesis, Swiss Federal Institute of Technology, Zrich 1979.
 */

public class DominoGame extends Process {
    public DominoGame(int nbrStones, double d, double g, double m, 
                      double x, double y, double z) {
        this.nbrStones = nbrStones;
        this.d = d; this.g = g; this.m = m;
        this.x = x; this.y = y; this.z = z;
    }
    
    public void actions() {
        dtMin = 1.0e-6; dtMax = 0.2;
        maxAbsError = maxRelError = 1.0e-4;
        ko = 1 - Math.pow(d - x, 2) / (z * z);
        km = m * g * 0.5 * Math.sqrt(x * x + z * z);
        kr = 1 - Math.sqrt(ko);
        phiPush = Math.asin((d - x) / z);
        theta = m * (x * x + z * z) / 3;
        activate(new Stone(z / 2 * 1.0e-5 / theta));
        waitUntil(new Condition() {
            public boolean test() { return fallingStones.empty(); }
        });
        v = d * (stones - 1) / time();    
    }

    class Stone extends Process {
        Stone(double initialOmega) {
            this.initialOmega = initialOmega;
        }
        
        public void actions() {
            into(fallingStones);
            stones++;
            phi = new Variable(0).start();
            omega = new Variable(initialOmega).start();
            fall = new StoneFall(this).start();
            if (stones < nbrStones) {
                waitUntil(new Condition() {
                    public boolean test() { return phi.state >= phiPush; }
                });
                activate(new Stone(ko * omega.state));
                omega.state *= kr;
            }
            waitUntil(new Condition() {
                public boolean test() { return phi.state >= Math.PI/2; }
            }); 
            out();
            fall.stop();
            phi.stop();
            omega.stop();    
        }
        
        double initialOmega;
        Variable phi, omega;
        Continuous fall;
    }
    
    class StoneFall extends Continuous {
        StoneFall(Stone s) {
            this.s = s;
        }
        
        public void derivatives() {
            s.omega.rate = km * Math.sin(s.phi.state) / theta;
            s.phi.rate = s.omega.state;
        }
        
        Stone s;
    }
    
    int nbrStones;
    double d, g, m, x, y, z;
     
    double v, ko, km, kr, phiPush, theta;
    int stones;
    Head fallingStones = new Head();
        
    public static void main(String[] args) {
        Function velocity = new Function() {
            public double f(double x) {
                DominoGame game = new DominoGame(55, x, 9.81, 0.01, 0.008, 0.024, 0.046);
                activate(game);
                return game.v;
            }
        };
        Function.Pair p = velocity.maximum(0.008, 0.008 + 0.046, 0.001);
        Format.print(System.out, 
               "The maximum chain velocity %5.3f m/s\n", p.x);
        Format.print(System.out, 
               "is reached with a distance of %6.4f m between stones\n", p.y);
    }
}

public abstract class Function {
    public abstract double f(double x);
    
    Pair maximum(double a, double b, double tol) {
        double f = (Math.sqrt(5) - 1) / 2;
        double x1 = b - f * (b - a);
        double y1 = f(x1);
        double x2 = a + f * (b - a);
        double y2 = f(x2);    
        
        while (b - x1 > tol) {
            if (y1 >= y2) {
                b = x2; x2 = x1; y2 = y1; x1 = b - f * (b - a);
                y1 = f(x1);
            } else {
                a = x1; x1 = x2; y1 = y2; x2 = a + f * (b - a);
                y2 = f(x2);
            }
        }
        if (y1 >= y2) 
            return new Pair(x1, y1);
        return new Pair(x2, y2);
    }
    
    public class Pair {
        Pair(double x, double y) {
            this.x = x;
            this.y = y;
        }
        
        public double x, y;
    }
}
