import jDisco.*;
import jDisco.Process;

/*
 * This example is a simulation of a fire station. The model is formulated 
 * by R. W. Sierenberg, Delft University, the Netherlands.
 *
 * A small city owns a fire station with three fire engines. Fire alarms are 
 * given randomly at exponentially distributed intervals with a mean of six hours.
 *
 * Each house on fore contains a certain amount of inflammable material. 
 * When a fire is discovered, it already has a certain size. 
 * The fire increases with a rate that is proportional to its size as no 
 * extinguishing activity takes place.
 *
 * At the moment an alarm is given, one fire engine, if it is available, will 
 * be sent to the fire. When a fire engine reaches the fire and finds out that 
 * its capacity is smaller than the rate with which the fire increases, it will 
 * request assistance by sending a second alarm for the same fire. The fire engine 
 * returns to the fire station after the fire is put out or when all inflammable 
 * material is consumed.
 *
 * Class HouseOnFire describes the discrete events associated with a house on fire: 
 * alarm call and fire termination.
 *
 * Class Burning describes, through differential equations, the continuous process 
 * associated with a house on fire.
 *
 * Class FireEngine defines the fire engines.
 * 
 * An Incendiary-object sets houses on fire.
 * 
 * The simulation period is one month.
 *
 * Reference:
 *     
 *    Sierenberg, R. W.
 *    Combined Discrete and Continuous Simulation with PROSIM
 *    Proceedings Simulation '77 (M. H. Hamza, editor).
 */

public class FireFighting extends Process {
    Head fireStation = new Head();
    Head alarmQ = new Head();
    static final int oneHour = 60;
    static final int oneMonth = 30 * 24 * oneHour;
    Random rand = new Random(54521);
    Histogram percDamage = new Histogram("Perc. damage", 0, 100, 20);

    public void actions() {
        dtMin = 0.0001; dtMax = oneHour; 
        maxRelError = maxAbsError = 0.0001;
        for (int i = 1; i <= 4; i++)
            new FireEngine(10).into(fireStation);
        activate(new Incendiary());
        hold(oneMonth);
        System.out.println("qLength = " + alarmQ.cardinal());
        // percDamage.report();
    }
    
    class Burning extends Continuous {
        Burning(HouseOnFire house) {
            this.house = house;
        }
        
        public void derivatives() {
            house.size.rate   = house.c * house.size.state - house.extinguishRate;
            house.damage.rate = house.size.state;
        }
        
        HouseOnFire house;
    }
    
    class HouseOnFire extends Process {
        double c, material, extinguishRate, travelTime;
        Variable size, damage;
        Burning fire;
        
        public void actions() {
            c = rand.uniform(0.01, 0.06);
            material = rand.uniform(100, 600);
            travelTime = rand.uniform(5, 15);
            size = new Variable(rand.uniform(0, 5)).start();
            damage = new Variable(size.state).start();
            fire = new Burning(this);
            fire.start();
            new Alarm(this).into(alarmQ);
            activate((Process) fireStation.first());
            waitUntil(new Condition() {
                public boolean test() {
                    return size.state <= 0 || damage.state >= material;
                }
            });
            percDamage.update(damage.state * 100 / material);
            fire.stop();
            damage.stop();
            size.stop();
        }
    }
    
    class FireEngine extends Process {
        FireEngine(double capacity) {
            this.capacity = capacity;
        }
        
        public void actions() {
            while (true) {
                Alarm alarm = (Alarm) alarmQ.first();
                if (alarm != null) {
                    alarm.out();
                    HouseOnFire house = alarm.house;
                    if (house.fire.isActive()) {
                        out();
                        hold(house.travelTime);
                        if (house.size.rate > capacity) {
                            alarm.into(alarmQ);
                            activate((Process) fireStation.first());
                        }
                        house.extinguishRate += capacity;
                        while (house.fire.isActive())
                            hold(5);
                        hold(house.travelTime);
                        into(fireStation);
                    }
                } else 
                    passivate();
            }
        }
        
        double capacity;
    }
    
    public class Incendiary extends Process {
        public void actions() {
            while (true) {
                hold(rand.negexp(1 / (6.0 * oneHour)));
                activate(new HouseOnFire());
            }
        }
    }
    
    class Alarm extends Link {
        Alarm(HouseOnFire house) {
            this.house = house;
        }
        
        HouseOnFire house;
    }
    
    public static void main(String[] args) {
        for (int i = 1; i <= 10; i++) {
        activate(new FireFighting());
        System.out.println(Thread.activeCount());
        }
    }
}
