import jDisco.*;
import jDisco.Process;

/* 
 * Simulation of a block resting on a rough surface, attached to a spring
 * to a point that can be moved, thus applying a force to the block via the
 * compression or extension of the spring.
 *
 * The model has been taken from
 *
 *     Advanced Continuous Simulation Language (ACSL)
 *     Reference Manual, 1999, (p. A-102 - A-115)
 *     http://www.ACSL.sim.com
 */

public class FrictionTest extends Process {
    static final double xfa = 0.05, freqf = 0.40;
    static final double ksp = 100.0, kbvd = -0.50, mb = 1.0;
    static final double xbdic = 0.0, xbic = 0.0;
    static final double kbbf = 2.50, kbsf = 2.00;
    
    Variable xb, xbd;
    boolean stukb;
    double xf, sumfb, fbcf, fbsf, xbdd;
    double fib;
    
    public void derivatives() {
        xf = xfa * Math.sin(2.0 * Math.PI * freqf * time());
        sumfb = ksp * (xf - xb.state) + kbvd * xbd.state;
        fbcf = stukb ? -sumfb : fbsf;
        xbdd = (sumfb + fbcf) / mb;
        xbd.rate = xbdd;
        xb.rate = xbd.state;
        fib = stukb ? Math.abs(sumfb) - kbbf : xbd.state;
    }
    
    public void actions() {
        xb   = new Variable(xbic).start();
        xbd  = new Variable(xbdic).start();
        start();
        final Graph graph_xb   = new Graph(" xb ");
        final Graph graph_xf   = new Graph(" xf ");
        final Graph graph_fib  = new Graph(" fib ");
         class StateReporter extends Reporter {    
            public void actions() {
                graph_xb.add(xb);
                graph_xf.add(xf);
                graph_fib.add(fib);
            }
        }
        new StateReporter().setFrequency(0.020).start();
        activate(new Stick());
        hold(5);
        graph_xb.show(); 
        graph_xf.show();
        graph_fib.show();
    }
    
    class Stick extends Process {    
        public void actions() {
            stukb = xbd.state == 0.0;
            while (true) {
                waitUntil(new Condition() {
                    public boolean test() {
                        return Math.abs(fib) >= 1e-7;
                    }
                });
                if (fib <= 0) {
                    waitUntil(new Condition() {
                        public boolean test() {
                            return fib > 0;
                        }
                    });
                } else {
                    waitUntil(new Condition() {
                        public boolean test() {
                            return fib <= 0;
                        }
                    });
                }
                stukb = !stukb && Math.abs(sumfb) < kbbf;
                fbsf = stukb ? 0.0 : (-sumfb >= 0 ? 1 : -1) * Math.abs(kbsf); 
                xbd.state = 0.0;
            }
        }
    }
    
    public static void main(String[] args) {
        activate(new FrictionTest());
    }
}
