import jDisco.*;
import jDisco.Process;

/*
 * Simulation of the heating of steel ingots in a furnace.
 *
 * In the production of steel, ingots are heated to a desired temperature in a
 * funace of limited capacity.
 *
 * The ingots (class Ingot) each have their own particular temperature when they
 * arrive. If the furnace is full, an arriving ingot will have to wait in a queue.
 * When space is available, an ingot will be added to the furnace and heated until
 * it has reached the desired temperature. Subsequently it leaves the system
 * providing space in the furnace.
 *
 * Thermodynamic differential equations (class Thermodynamics) describe the
 * continuous temperature cahnges of ingots waiting to come into the furnce,
 * of those inside the furnace, and of the furnace itself.
 * 
 * The model has been taken from
 * 
 *    Sim, R.  
 *    CADSIM. User's Guide and Reference Manual  
 *    Imperial College. Publ. no. 75/23. London 1975
 */

public class FurnaceSimulation extends Process {
    Head furnace = new Head(), queue = new Head();
    Variable furnaceTemp;
    static final Random rand1 = new Random(7913), rand2 = new Random(23);
    static final double simPeriod = 100;
    Graph furnaceTempGraph = new Graph("Furnace temperature");
    Graph furnaceGraph = new Graph("Ingots in furnace");

    public void actions() {
        dtMin = 0.001; dtMax = 10.0; maxRelError = 0.01;
        furnaceTemp = new Variable(400).start();
        new ThermoDynamics().start();
        activate(new IngotGenerator());
        new StateReporter().setFrequency(0.0).start();
        hold(simPeriod);
        furnaceGraph.show();
        furnaceTempGraph.show();
    }
    
    class Ingot extends Process {
        Variable temp = new Variable(rand1.negexp(1 / 200.0));
        
        public void actions() {
            double arrivalTime = time();
            temp.start();
            if (furnace.cardinal() == 10)
                wait(queue);
            hold(4.612);
            into(furnace);
            waitUntil(new Condition() {
                public boolean test() { return temp.state > 380; }
            });
            out();
            temp.stop();
            activate((Process) queue.first());
        }
    }
    
    class ThermoDynamics extends Continuous {
        double sum, loss;
        
        public void derivatives() {
            Ingot ing;
            for (ing = (Ingot) queue.first(); ing != null; ing = (Ingot) ing.suc()) 
                ing.temp.rate = (25 - ing.temp.state) / 7;
            int n = 0;
            double sum = 0;
            for (ing = (Ingot) furnace.first(); ing != null; ing = (Ingot) ing.suc()) {
                ing.temp.rate = (furnaceTemp.state - ing.temp.state) / 7;
                n++;
                sum += ing.temp.state;
            }
            double loss = n > 0 ? n * (furnaceTemp.state - sum / n) / 3  : 0;
            furnaceTemp.rate = (500 - furnaceTemp.state) - loss;
        }
    }
    
    class IngotGenerator extends Process {
        public void actions() {
            while (true) {
                activate(new Ingot());
                hold(rand2.negexp(1 / 4.2));
            }    
        }
    }
    
    class StateReporter extends Reporter {
        public void actions() {
            furnaceTempGraph.add(furnaceTemp);
            furnaceGraph.add(furnace.cardinal());
        }
    }
    
    public static void main(String[] args) {
        activate(new FurnaceSimulation());
    }
}



