import jDisco.*;
import jDisco.Process;

/*
 * Simulation of a harmonic oscilator,
 *
 *     y'' + 4^2y = 0 
 *
 * with initial conditions of y(0) = 1, y'(0) = 0.
 * 
 * The exact solution is y(t) = cos 2t.
 *
 * The program computes for eeach built-in variable step-size integration 
 * method a summary of the absolute difference between the exact solution 
 * and numerical solution as a function of error-bound. The number of steps 
 * is also listed. 
 */

public class HarmonicOscilator extends Process {
    public HarmonicOscilator(String methodName, double maxError) {
        integrationMethod(methodName);
        maxAbsError = maxRelError = maxError;
    }

    Variable y, yd;
    
    public void derivatives() { 
        yd.rate = -4.0 * Math.pow(Math.PI, 2) * y.state;
        y.rate = yd.state;
    } 
    
    public void actions() {
        y = new Variable(1).start();
        yd = new Variable(0).start();
        start();
        ErrorReporter r = (ErrorReporter) new ErrorReporter().start();   
        hold(1.0);
        Format.print(System.out, "%0.7f", maxAbsError);
        Format.print(System.out, "%17.2e", r.errorSum / r.steps);
        Format.print(System.out, "%10d\n", r.steps);
    }

    class ErrorReporter extends Reporter {
        int steps;
        double errorSum;

        public void actions() {
            if (dt() > 0) {
                steps++;
                errorSum += Math.abs(y.state - Math.cos(2 * Math.PI * time()));
            }
        }
    }

    static void solve(String methodName) {
        System.out.println("Solution error using " + methodName +
            " as a function of\n" +
            "     maxAbsError and maxRelError\n");
        System.out.println("Error-bound    Average error     Steps");
        for (int i = 7; i > 0; i--)
            activate(new HarmonicOscilator(methodName, Math.pow(10, -i)));
        System.out.println("--------------------------------------\n");
    }
    
    public static void main(String[] args) {
        String[] method = {"AdamsBashforth", "RKDP45", "RKE", 
                           "RKF45", "RKN34", "RKV56"};
        for (int i = 0; i < method.length; i++)
            solve(method[i]);
    }
}