import jDisco.*;
import jDisco.Process;

/*
 * This example program simulates a limit cycle in the XY plane described by
 * the equations
 *
 *    x' =  y + (kx(1-x^2-y^2)/(sqrt(x^2 + y^2))
 *    y' = -x + (ky(1-x^2-y^2)/(sqrt(x^2 + y^2))
 *
 * where
 *
 *    x(0) = xz
 *    y(0) = yz
 *
 * The limit cycle is a circle of radius 1.0; that is, no matter what initial
 * conditions are imposed on x and y (except x = y = 0), (x^2 + y^2) -> 1 as
 * t -> infinity.
 *
 * The model has been taken from
 *
 *    Advanced Continuous Simulation Language (ACSL)
 *    Reference Manual, 1999, (p. A-34 - A-40)
 *    http://www.ACSL.sim.com
 */

public class LimitCycle extends Process {
    static final double xz = 0.5, yz = 1.0;
    static final double  k = 0.2;
    Variable x, y;
    
    public void derivatives() {
        maxRelError = 1e-10;
        double x2 = Math.pow(x.state, 2);
        double y2 = Math.pow(y.state, 2);
        double sq = Math.sqrt(x2 + y2);
        double kk = k * (1 - x2 - y2) / sq; 
        x.rate =  y.state + x.state * kk;
        y.rate = -x.state + y.state * kk;
    } 
    
     public void actions() {
         x = new Variable(xz).start();
         y = new Variable(yz).start();
         start(); 
         final Graph graph = new Graph(" y / x ");
         class StateReporter extends Reporter {    
            public void actions() {
                graph.add(x, y);
            }
        }
        new StateReporter().setFrequency(0.2).start();
        hold(10);
        graph.show();
    }
    
    public static void main(String[] args) {
        long startTime = System.currentTimeMillis();
        activate(new LimitCycle());
        long stopTime = System.currentTimeMillis();
        System.out.println("Time used = " + 
            (stopTime - startTime) / 1000.0 + " seconds");
    }
}


