import jDisco.*;
import jDisco.Process;

/*
 * Simulation of a pilot ejection system.
 *
 * A pilot ejection system, when activated, causes the pilot and 
 * his seat to travel along rails at a specified exit velocity (ve) 
 * at an angle (thetad) backward from vertical. After traveling a 
 * vertical distance (y1), the seat becomes disengaged from its 
 * mounting rail. At this point a second phase begins during which 
 * the pilot's trajectory is influ-enced by the force of gravity 
 * and atmospheric drag.
 *
 * The two phases are continuous. 
 *
 * The first phase is described by two differen-tial equations. 
 *
 * The second phase is described by four differential equations. 
 *
 * A discrete event separates the two phases.
 *
 * The aim of the simulkation is to determine how large the maximum 
 * velocity of the aircraft (va) may be, as a function of the height 
 * above sea level (h), in order for a safe ejection. An ejection is 
 * said to be "safe" if the ejection seat clears the vertical stabilizer 
 * of the aircraft, which is 30 feet behind the cockpit, at a distance 
 * of at least 20 feet.
 *
 * Reference:
 *
 *     SCi Simulation Software Committee  
 *     The SCi Continuous System Simulation Language (CSSL)
 *     SIMULATION, Vol. 9, December 1967, pp. 281-303.
 */

public class PilotEjection extends Process {
    public PilotEjection(double vA, double h) {
        this.vA = vA; this.h = h;
    }
    
    public void actions() {
        dtMax = 1; maxRelError = 0.00001;
        density = new Table();
        density.add(    0, 2377e-6); density.add( 1000, 2308e-6); 
        density.add( 2000, 2241e-6); density.add( 4000, 2117e-6); 
        density.add( 6000, 1987e-6); density.add(10000, 1755e-6); 
	    density.add(15000, 1497e-6); density.add(20000, 1267e-6); 
	    density.add(30000,  891e-6); density.add(40000,  587e-6); 
	    density.add(50000,  364e-6); density.add(60000, 2238e-7);
	    rhoP = density.value(h) * cD * s;
	    vx = vA - vE * Math.sin(thetaE);
	    vy = vE * Math.cos(thetaE);
	    x = new Variable(0).start();
	    y = new Variable(0).start();
	    v = new Variable(Math.sqrt(vx * vx + vy * vy));
	    theta = new Variable(Math.atan(vy / vx));
	    phase = new Phase1().start();
	    waitUntil(new Condition() {
	        public boolean test() { return y.state >= y1; }
	    });
	    phase.stop();
	    phase = new Phase2().start();
	    v.start();
	    theta.start();
	    waitUntil(new Condition() {
	        public boolean test() { return x.state <= -30; }
	    });
	    safe = y.state >= 20;
    }
    
    class Phase1 extends Continuous {
        public void derivatives() {
            x.rate = v.state * Math.cos(theta.state) - vA;
            y.rate = v.state * Math.sin(theta.state);
        }
    }
    
    class Phase2 extends Phase1 {
        public void derivatives() {
            super.derivatives();
            v.rate = (-0.5 * rhoP * (v.state * v.state)) / m - g * Math.sin(theta.state);
            theta.rate = -g * Math.cos(theta.state) / v.state;
        }
    }
    
    double vA, h;
    boolean safe;
    static final double m = 7, g = 32.2, cD = 1, s = 10, y1 = 4, 
                        vE = 40, thetaD = 15, thetaE = thetaD / 57.3;
    double vx, vy, rhoP;
    Variable x, y, v, theta;
    Continuous phase;
    static Table density;
    
    public static void main(String[] args) {
        double h = 0;
        Graph graph = new Graph("h / vA");
        for (double vA = 100; vA <= 900; vA += 50) {
            while (true) {
                PilotEjection pe = new PilotEjection(vA, h);
                activate(pe);
                if (pe.safe)
                    break;
                h += 500;
            } 
            graph.add(vA, h);
        }
        density.print("Density");
        density.show("Density");
        graph.show();     
    }
}
