import jDisco.*;
import jDisco.Process;

/*
 * Simulation of a predator-prey system.
 *
 * An object of class PredatorPreySystem governs the simulation. 
 * It starts two state variables and one continuous process, and 
 * finally suspends itself for a period of 100 time units 
 * (the simulation period).
 *
 * The system is known to be cyclic. The period can be determined 
 * through the waitUntil method by computing the time interval between 
 * two consecutive maximum points of a state variable, say prey.
 */

public class PredatorPreySystem extends Process {
    Variable predator, prey;
    Graph graph = new Graph(" Prey / Predator "); 
    
     public void actions() {
         dtMin = 1.0e-5; dtMax = 1; 
         maxAbsError = 0; maxRelError = 1.0e-5;
         predator  = new Variable(1000).start();
         prey = new Variable(100000).start();
         new Dynamics().start();
         class StateReporter extends Reporter {    
            public void actions() {
                graph.add(predator, prey);
            }
        }
        new StateReporter().setFrequency(0.1).start();
        waitUntil(new Condition() {
            public boolean test() { return prey.rate > 0; }
        });
        waitUntil(new Condition() {
            public boolean test() { return prey.rate <= 0; }
        });
        double cycleStartTime = time();
        waitUntil(new Condition() {
            public boolean test() { return prey.rate > 0; }
        });
        waitUntil(new Condition() {
            public boolean test() { return prey.rate <= 0; }
        });
        graph.show();
        Format.print(System.out, "Cycle time = %0.2f\n", time() - cycleStartTime);
    }
    
    class Dynamics extends Continuous {
        public void derivatives() {
            predator.rate  = (-0.3 + 3.0e-7 * prey.state) * predator.state; 
            prey.rate = (0.3 - 3.0e-4 * predator.state)  * prey.state; 
        }
    }
    
    public static void main(String[] args) {
        activate(new PredatorPreySystem());
    }
}
