import jDisco.*;
import jDisco.Process;

/*
 * Simulation of the temperature distribution alang a radiating fin. 
 *
 * The only practical way of rejecting heat from a power plant operating
 * in outher space is by thermal radiation. The working fluid of the power
 * plant passes through tubes. The space between tubes is filles by a fin
 * of rectangular cross secstion.
 * 
 * The aim of the simulation is to determine temperature profile for 
 * various tube spacings, fin thickness, fin material, etc. 
 * The temperature profile in turn can be used to to calculate the
 * effeciency of the radating system.
 *
 * The model has been taken from
 *
 *     Advanced Continuous Simulation Language (ACSL)
 *     Reference Manual, 1999, (p. A-20 - A-33)
 *     http://www.ACSL.sim.com
 */

public class RadiatingFin extends Process {
    static final double sg = 1.73e-9, ep = 0.9, h = 0.00125,
                             k = 25.0, ts = 0.0;
    static final double tstop = 0.25, tz = 2000.0;
    
    Variable t, dtdx;
    double dtdxz;
    
    public RadiatingFin(double dtdxz) {
        this.dtdxz = dtdxz;
    }
    
    public void derivatives() {
        dtdx.rate = sg * ep * (Math.pow(t.state, 4) - Math.pow(ts, 4)) / (k * h);
        t.rate = dtdx.state;
    }
    
    public void actions() {
        t = new Variable(tz).start();
        dtdx = new Variable(dtdxz).start();
        start();
        waitUntil(new Condition() {
            public boolean test() {
                return time() >= tstop || t.state >= 1.01 * tz || t.state < 0;
            }
        });
    }
    
    public static void main(String[] args) {
        double dtdxz = -20000, error = 0.2, gain = 0.07;
        for (int n = 0; n < 100; n++) {
            RadiatingFin rf = new RadiatingFin(dtdxz);
            activate(rf);
            if (Math.abs(rf.dtdx.state) < error) {
                System.out.println("t = " + rf.t.state + ", dtdxz = " + rf.dtdxz);
                return;
            }
            dtdxz -= gain * rf.dtdx.state;
        }
    }
}
