import jDisco.*;
import jDisco.Process;

/*
 * This example has to do with the launch of a three-stage rocket from the 
 * earth's surface. 
 *
 * The model has both continuous and discrete elements. 
 *
 * During the rocket's flight well-known physical laws govern its motion, and 
 * its mass decreases continuously as a result of the expulsion of burnt fuel. 
 * However, when a stage separates from the rest of the rocket, an instantaneous 
 * change of the rocket's mass and acceleration takes place.
 * 
 * The continuous motion of the rocket is described in class RocketMotion. 
 * The change in mass takes place at a constant rate, massFlow. There are 
 * three forces acting upon the rocket: thrust, drag and gravity. 
 * thrust is the force generated by the expulsion of burnt fuel, 
 * drag is the air resistance, and gravity is the earth's gravitational 
 * attraction on the rocket.
 *
 * The discrete changes (that is, the separation of stages) are described in the 
 * discrete main process (class ThreeStageRocket).
 *
 * The rocket's launch begins when the three Variable-objects mass, velocity and 
 * altitude are generated and started together with an object of class RocketMotion. 
 * Each time a stage separates, discrete changes in mass, massFlow, flowVelocity 
 * and area take place. These events are scheduled to occur by the method hold. 
 */

public class ThreeStageRocket extends Process {
    Variable mass, velocity, altitude;
    double thrust, drag, gravity, massFlow, flowVelocity, area;
        
    class RocketMotion extends Continuous {
        public void derivatives() {
            mass.rate = -massFlow;    
            thrust = massFlow * flowVelocity;
            drag = area * 0.00119 * Math.exp(-altitude.state / 24000) * 
                Math.pow(velocity.state, 2);
            gravity = mass.state * 32.17 / Math.pow(1 + altitude.state / 20908800, 2);    
            velocity.rate = (thrust - drag - gravity) / mass.state;
            altitude.rate = velocity.state;
        }
    }
    
    public void actions() {    
        final Graph massGraph = new Graph("Mass");
        final Graph velocityGraph = new Graph("Velocity");
        final Graph altitudeGraph = new Graph("altitude");
        new Reporter() {    
            public void actions() {
                massGraph.add(mass);
                velocityGraph.add(velocity);    
                altitudeGraph.add(altitude);
            }
        }.start();
        
        dtMin = 0.00001; dtMax = 100; 
        maxAbsError = maxRelError = 0.00001;
        
        // first stage
        massFlow = 930; flowVelocity = 8060; area = 510;
        mass     = new Variable(189162).start();
        velocity = new Variable(0).start();
        altitude = new Variable(0).start();
        new RocketMotion().start();
        hold(150);
        
        // second stage
        mass.state = 40342; massFlow = 81.49; flowVelocity = 13805; area = 460;
        hold(359);
        
        // third stage
        mass.state =  8137; massFlow = 14.75; flowVelocity = 15250; area = 360;
        hold(479);
        massGraph.show();
        velocityGraph.show();
        altitudeGraph.show();
    }
    
    public static void main(String[] args) {
        activate(new ThreeStageRocket());
    }
}
        
